#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "types.h"
#include "rand.h"
#include "recipes.h"

#define pi 3.141592653589793238462643
#define epsilon 1e-10		/* thickness of surfaces */
#define infinity 6e66

/* --------------------------------------------------- */
/*                                                     */
/* this is the code for some geometric routines        */
/* related to calculations of Minkowski functionals    */
/* --------------------------------------------------- */
void freetriple(triple*);
void emptylist(triple**);
void sortintolist(triple**,section*,triple*);
void triplepoint(triple**,section*,body*,pattern*);
double lhuilier(double**);
void getnormal(double*,double*,body*);

double curvelength(section*,triple*,pattern*);
double curvelet(double,double,section*,pattern*);
double balltriangle(section*,body*,pattern*,double,double);
double chipos(double, section*,body*);
double chiphi(double, section*,body*,pattern*);
double triangle(double,double, section*,body*,pattern*);
double narrowtriangle(double, section*,body*,pattern*);
void curvepoint(double*,section*,double);
void curvetangent(double*,section*,double);
void sectioncurve(section**,body*,body*);
boolean intersection(body*,body*);

double**scanpoints(body*,int*,double*);

boolean irrelevant(body*,pattern*);
boolean involume(pattern*,double*);
boolean ontriple(body*,body*,body*,pattern*,double*);
boolean oncurve(body*,body*,pattern*,double*);       
boolean onsurface(body*,pattern*,double*);
boolean inbody(body*,double*);
boolean outbody(body*,double*);
double bodydist(body*,double*);

void orthsystem(double*,double*,double*);
void vprod(double*,double*,double*);
double dist(double*,double*);
double dvec(double*,double*,double*);
double sprod(double*,double*);
double saferatan2(double,double);
double saferacos(double);
double saferasin(double);
double safersqrt(double);

void outputlist(triple*);

/* --------------------------------------------- */
/*                                               */
/* these routines relate to the calculation      */
/* of triple points of three bodies              */
/* --------------------------------------------- */
/* free memory of a single triple point */
void freetriple(triple*t)
{
  free((void*)t->parent); free((void*)t->x); free((void*)t); 
}
/* empty list of triple points */
void emptylist(triple**trip)
{
  triple*t;

  /* free anything the list contains */
  for(t=*trip; t; t=t->next) freetriple(t);
  *trip = (triple*)NULL; 
}
/* sort a new triple point into an existing list of triple points */
void sortintolist(triple**trip, section*s, triple*new)
{
  triple*t,*ring=(triple*)NULL;

  /* if on a circle, close ring list */
  switch(s->type) {
  case circ : ring=new; break;
  case line : ring=(triple*)NULL; break;
  case nosect : break;
  }

  if(*trip&&new->par>(*trip)->par) {
    /* find predecessor in list */
    for(t=*trip;t->next&&new->par>=t->next->par;t=t->next);
      
    /* set pointers */
    new->prev = t;
    new->next = t->next;
    if(t->next) t->next->prev=new; else (*trip)->prev=ring;
    t->next=new;
  }
  else {
    /* put at beginning of list */
    new->next = *trip;
    if(*trip)new->prev = (*trip)->prev, (*trip)->prev = new;
    else new->prev = ring;
    *trip = new;
  }
  
}
/* calculate the triple point of three bodies */
/* and add it to triple point list */
/* the intersection line of the first two is already known */
void triplepoint(triple**tr, section*s,body*b,pattern*pat)
{
  triple*new;
  int i,j, ntrip=0, nwin=0, sign;
  double trip[2], x[3], euler;
  double phi,psi=0, d[3],proj[2], dista,denom,rad,cosine=0, discr;
  static double**normal=(double**)NULL;

  if(!normal) normal=dmatrix(0,2,0,2);

  /* calculate all possible triple points; store parameters on curve */
  switch(s->type) {
  case circ :
    switch(b->type) {
    case ball :
    case hole :
      /* calculate axis vector pointing towards third point */
      dista=dvec(d,b->par[0],s->par[0]);
      for(i=0;i<2;i++) proj[i]=sprod(d,s->par[i+1]);
      /* calculate angular separation of triple points */
      if(proj[0]==0&&proj[1]==0)
	cosine=infinity;
      else {
	psi=saferatan2(proj[1],proj[0]);
	denom=2*(cos(psi)*proj[0]+sin(psi)*proj[1]);
	rad=sprod(s->par[1],s->par[1]);
	cosine=(b->par[1][0]*b->par[1][0]-rad-dista*dista)/denom;
      }
      break;
    case wall :
      /* calculate axis vector pointing towards third point */
      for(i=0; i<2; i++) proj[i] = sprod(b->par[1],s->par[i+1]);
      psi = saferatan2(proj[1],proj[0]);
      /* calculate angular separation of triple points */
      denom = cos(psi)*proj[0]+sin(psi)*proj[1];
      rad = sprod(b->par[0],b->par[1])-sprod(s->par[0],b->par[1]);
      cosine = rad/denom;
      break;
    case nobody : break;
    }

    /* calculate triple points, if existent */
    if(1+cosine>epsilon&&1-cosine>epsilon) {
      phi=saferacos(cosine); ntrip=2; trip[0]=psi+phi; trip[1]=psi-phi;
      for(i=0;i<2;i++) {
	while(trip[i]<0) trip[i]+=2*pi; while(trip[i]>= 2*pi) trip[i]-=2*pi;
      }
    }
    break;
  case line :
    switch(b->type) {
    case ball :
    case hole :
      /* discriminant of equation to solve */
      dvec(d,b->par[0],s->par[0]); dista = sprod(d,s->par[1]);
      discr = dista*dista-sprod(d,d)+b->par[1][0]*b->par[1][0];
      /* calculate triple point */
      if(discr>0) {
	ntrip=2; 
	trip[0]=-dista+safersqrt(discr); 
	trip[1]=-dista-safersqrt(discr);
      }
      break;
    case wall :
      /* calculate parameter on line */
      denom = sprod(s->par[1],b->par[1]);
      if(fabs(denom) > epsilon) {
	dista = sprod(b->par[0],b->par[1])-sprod(s->par[0],b->par[1]);
	ntrip = 1; trip[0] = dista/denom;
      }
      break;
    case nobody : break;
    }
    break;
  case nosect : break;
  }

  /* if relevant for pattern, store in list */
  for(i=0;i<ntrip;i++) {
    curvepoint(x,s,trip[i]); 
    if(ontriple(s->parent[0],s->parent[1],b,pat,x)) {
      /* make a new triple point */
      new = (triple*)calloc(1,sizeof(triple));
      new->x = (double*)calloc(3,sizeof(double)); 
      new->parent = (body**)calloc(3,sizeof(body*));

      for(j=0; j<3; j++) new->x[j] = x[j];
      new->par = trip[i];
      new->parent[0] = s->parent[0]; 
      new->parent[1] = s->parent[1]; 
      new->parent[2] = b;

      /* and sort it into the list */
      sortintolist(tr,s,new);
      
      /* evaluate contribution to functionals */
      if(s->parent[0]->n<s->parent[1]->n && s->parent[1]->n<b->n) {
	/* get normals and sign */
	for(j=0, nwin=0; j<3; j++) {
	  getnormal(normal[j],x,new->parent[j]);
	  if(new->parent[j]->type != ball) nwin++;
	}
	if(nwin==1) sign=-1; else sign=+1;
	euler=sign*lhuilier(normal)/(4*pi);

	/* share among bodies */
	for(j=0; j<3; j++) new->parent[j]->v[3] += euler/3;
      }
    }
  }

}
/* calculate spherical excess of triangle spanned by normals */
double lhuilier(double**n)
{
  double l[3], s[3], c, a[3];
  int i;

  /* lengths and scalar products; angles; quarter circumference*/
  for(i=0; i<3; i++) l[i] = sprod(n[i],n[i]), s[i] = sprod(n[i],n[(i+1)%3]);
  for(i=0; i<3; i++) a[i] = saferacos(s[i]/safersqrt(l[i]*l[(i+1)%3]));
  c = (a[0]+a[1]+a[2])/4;

  /* calculate spherical excess */
  return 4*atan(safersqrt(tan(c)*tan(c-a[0]/2)*tan(c-a[1]/2)*tan(c-a[2]/2)));
}
/* get normal n of body b in point x */
void getnormal(double*n, double*x, body*b)
{
  int i;

  switch(b->type) {
  case ball :
  case hole : 
    for(i=0; i<3; i++) n[i] = (x[i]-b->par[0][i])/b->par[1][0];
    break;
  case wall :
    for(i=0; i<3; i++) n[i] = b->par[1][i];
    break;
  case nobody : break;
  }
} 

/* --------------------------------------------- */
/*                                               */
/* routines for intersection of two bodies       */
/* test, calculation & measuring                 */
/* --------------------------------------------- */
/* calculate length of a given intersection curve within a pattern */
/* triple point list is already set */
double curvelength(section*s,triple*trip,pattern*pat)
{
  int i,j;
  double length=0;
  triple*t;

  for(t=trip;t;t=t->next)
    if(t->prev) length+=curvelet(t->prev->par,t->par,s,pat);

  if(!trip&&s->type==circ) length+=curvelet(0,2*pi,s,pat);

  /* share contribution among parent bodies */
  if(length && s->parent[0]->n<s->parent[1]->n)
    for(i=2; i<=3; i++) 
      for(j=0; j<2; j++) s->parent[j]->v[i] += length*s->v[i]/2;

  return length;
}
/* calculate the length of a `curvelet' and return it */
/* if on a flat surface, also calculate the contribution to the area */
double curvelet(double lo, double hi, section*s,pattern*pat)
{
  int i;
  double delta,t,x[3],area=0,chi=0,rad;
  boolean test;

  /* test middle of curve */
  delta=hi-lo; if(lo>hi) delta+=2*pi; t=lo+delta/2;
  curvepoint(x,s,t);
  test=oncurve(s->parent[0],s->parent[1],pat,x);

  if(test&&s->parent[0]->n<s->parent[1]->n) for(i=0;i<2;i++) {
    switch(s->parent[i]->type) {
    case hole :
    case ball : 
      rad=s->parent[i]->par[1][0];
      chi=balltriangle(s,s->parent[i],pat,lo,lo+delta);
      s->parent[i]->v[1]+=chi*4*pi*rad*rad; 
      break;
    case wall : 
      switch(s->type) {
      case circ :
	area=romberg(narrowtriangle,s,s->parent[i],pat,lo,lo+delta, "curvelet-narrowtringle");
	break;
      case line :
	area=(hi-lo)*triangle(lo,hi,s,s->parent[i],pat);
	break;
      case nosect : break;
      }
      s->parent[i]->v[1]+=area/6;
    case nobody : break;
    }
  }

  if(test) return delta; else return 0;
}
/* calculate integral over function chiphi in portions if necessary */
double balltriangle(section*s,body*b,pattern*pat,double lo, double hi)
{
  double chi=0, min,max, x=.1;

  min=lo; max=hi; if(max>2.5*pi) max-=2*pi, min-=2*pi;
  if(min<pi/2 && pi/2<max) {
    for(; chipos(min,s,b) < x*chipos(pi/2,s,b); min = x*min+(1-x)*pi/2)
      chi += romberg(chiphi,s,b,pat,min,x*min+(1-x)*pi/2, "balltriangle-chiphi-1");
    for(; chipos(max,s,b) < x*chipos(pi/2,s,b); max = x*max+(1-x)*pi/2)
      chi += romberg(chiphi,s,b,pat,x*max+(1-x)*pi/2,max, "balltriangle-chiphi-2");
  }
  if(chipos(min,s,b)<x*chipos(max,s,b))
    for(; chipos(min,s,b)<x*chipos(max,s,b); min = x*min+(1-x)*max)
      chi += romberg(chiphi,s,b,pat,min,x*min+(1-x)*max, "balltriangle-chiphi-3");
  if(chipos(max,s,b)<x*chipos(min,s,b))
    for(; chipos(max,s,b)<x*chipos(min,s,b); max = x*max+(1-x)*min)
      chi += romberg(chiphi,s,b,pat,x*max+(1-x)*min,max, "balltriangle-chiphi-4");
  chi += romberg(chiphi,s,b,pat,min,max, "balltriangle-chiphi-5");

  return chi;
}
double chipos(double t, section*s, body*b)
{
  double x[3], dx[3], rad=b->par[1][0], *m=b->par[0];
  double theta, phi, dphi, area;

  curvepoint(x,s,t); curvetangent(dx,s,t);
  theta = saferacos((x[2]-m[2])/rad); phi = saferatan2(x[1]-m[1],x[0]-m[0]);
  dphi = fabs(dx[1]*cos(phi)-dx[0]*sin(phi))/(rad*sin(theta));
  area = dphi * (1-cos(theta))/(4*pi);
    
  return area;
}
double chiphi(double t, section*s, body*b, pattern*pat)
{
  int sign;
  double x[3], dx[3], p[3], rad=b->par[1][0], *m=b->par[0];
  double theta, phi, dphi, area;
  boolean flag;
  body*c;

  curvepoint(x,s,t); curvetangent(dx,s,t);
  theta = saferacos((x[2]-m[2])/rad); phi = saferatan2(x[1]-m[1],x[0]-m[0]);
  dphi = fabs(dx[1]*cos(phi)-dx[0]*sin(phi))/(rad*sin(theta));
  area = dphi * (1-cos(theta))/(4*pi);

  p[0] = m[0] + rad*sin(theta*(1-1e-6))*cos(phi);
  p[1] = m[1] + rad*sin(theta*(1-1e-6))*sin(phi);
  p[2] = m[2] + rad*cos(theta*(1-1e-6));
    
  if(b == s->parent[0]) c = s->parent[1]; else c = s->parent[0];
  if(c->type==ball && b->type==ball) flag = outbody(c,p); 
  else flag = inbody(c,p);
  if(flag) sign = +1; else sign = -1;
  
  return sign * area;
}
/* evaluate the signed area of triangle spanned by two points on a curve */
/* and reference point of body -- must be a wall */
double triangle(double lo,double hi, section*s,body*b,pattern*pat)
{
  double delta;
  double *p=b->par[0],*n=b->par[1], x[3],y[3],u[3],v[3],m[3];
  double radius=1, height;
  boolean flag;
  int i, sign;

  /* radius for circle */
  switch(s->type) {
  case circ : radius = safersqrt(sprod(s->par[1],s->par[1])); break;
  case line : radius = 1; break;
  case nosect : break;
  }

  /* calculate the relevant points */
  /* x,y endpoints; u,v direction and normal of line in plane */
  curvepoint(x,s,lo); curvepoint(y,s,hi); curvetangent(u,s,(lo+hi)/2);
  vprod(v,u,n); height = fabs(sprod(x,v)-sprod(p,v));
  if(radius<height) {
      if(height < 1./infinity) fprintf(stderr, "height < 1/infty: %g\n", height);
      delta = .001*radius/height; 
  } else {
      delta = .01*radius;
  }
  for(i=0; i<3; i++) m[i] = delta*(p[i]-(x[i]+y[i])/2) + (x[i]+y[i])/2 ;

  /* determine sign */
  if(b == s->parent[0])		/* we have a neighbour list */
      if(onsurface(b,pat,m)) sign = +1; else sign = -1;
  else {			/* test'em all, shit - luckily this is rare */
      for(i=0, flag=true; i<pat->nwin && flag; i++) if(i != b->n) flag = inbody(&pat->b[i],x);
      if(flag) {
	  for(i=pat->nwin, flag=false; i<pat->n && !flag; i++) flag = inbody(&pat->b[i],x);
      }
      if(flag) sign = +1; else sign = -1;
  }

  return sign*radius*height/2;
}
/* same for very narrow triangle */
double narrowtriangle(double t,section*s,body*b, pattern*pat)
{
  double dt=.001;		/* something small */

  return triangle(t-dt,t+dt,s,b,pat);
}
/* calculate a point on the curve */
void curvepoint(double*x,section*s,double param)
{
  int i;

  switch(s->type) {
  case circ :
    for(i=0; i<3; i++) 
      x[i] = s->par[0][i]+cos(param)*s->par[1][i]+sin(param)*s->par[2][i];
    break;
  case line :
    for(i=0; i<3; i++)
      x[i] = s->par[0][i]+param*s->par[1][i];
    break;
  case nosect : break;
  }
}
/* calculate the tangent t to a curve s in point given by param */
void curvetangent(double*t,section*s,double param)
{
  int i;

  for(i=0; i<3; i++) {
    switch(s->type) {
    case circ : t[i] = -sin(param)*s->par[1][i]+cos(param)*s->par[2][i]; break;
    case line : t[i] = s->par[1][i]; break;
    case nosect : break;
    }
  }
}
/* calculate intersection curve */
void sectioncurve(section**s, body*a, body*b)
{
  int i,j;
  double *c1=a->par[0],*c2=b->par[0], *r1=a->par[1], *r2=b->par[1];
  double x[3], d, x1,x2, rad, cs, o1,o2,p1,p2;

  /* allocate memory, if necessary */
  if(!*s) {
    *s = (section*)calloc(1,sizeof(section));
    (*s)->parent = (body**)calloc(2,sizeof(body*));
    (*s)->par = (double**)calloc(4,sizeof(double*));
    for(i=0; i<3; i++) (*s)->par[i] = (double*)calloc(3,sizeof(double));
    (*s)->v = (double*)calloc(4,sizeof(double));
  }

  /* call the appropriate calculation routine */
  switch(a->type) {
  case ball : 
  case hole :
    switch(b->type) {
    case ball :
    case hole : (*s)->type = circ;

      /* center */
      d = dvec(x,c1,c2);
      x1 = .5+(*r1-*r2)*(*r1+*r2)/(2*d*d); /* relative distance from middles */
      x2 = .5+(*r2-*r1)*(*r1+*r2)/(2*d*d);
      for(i=0; i<3; i++) (*s)->par[0][i] = x2*c1[i] + x1*c2[i];
      
      /* axes */
      rad = safersqrt((x1<x2) ? (*r1-x1*d)*(*r1+x1*d) : (*r2-x2*d)*(*r2+x2*d));
      orthsystem((*s)->par[1],(*s)->par[2],x);
      for(i=0; i<3; i++) for(j=1; j<3; j++) (*s)->par[j][i] *= rad;

      /* local Minkowski functionals */
      x1 /= *r1/d; x2 /= *r2/d;	/* sines of opening angles */
      (*s)->v[2] = .5*rad*(saferasin(x1)+saferasin(x2)) / (3*pi);
      (*s)->v[3] = (x1+x2)/(4*pi);
      /* correct sign */
      if(a->type==b->type) (*s)->v[2] *= -1, (*s)->v[3] *= -1;

      break;
    case wall : (*s)->type = circ;
      
      /* center ; whole circle lies in wall */
      d = bodydist(b,c1);
      for(i=0; i<3; i++) (*s)->par[0][i] = c1[i] - d*r2[i];
      
      /* axes */
      rad = safersqrt((*r1-d)*(*r1+d));
      orthsystem((*s)->par[1],(*s)->par[2],r2);
      for(i=0; i<3; i++) for(j=1; j<3; j++) (*s)->par[j][i] *= rad;

      /* local Minkowski functionals */
      x1 = d/ *r1;		/* sine of opening angle */
      (*s)->v[2] = .5*rad*(pi/2+saferasin(x1)) / (3*pi);
      (*s)->v[3] = (1+x1)/(4*pi);

      break;

    case nobody : break;
    }
    break;
  case wall :
    switch(b->type) {
    case ball :
    case hole : sectioncurve(s,b,a); break;
    case wall : (*s)->type = line;
      
      /* some point on the line by orthogonal components*/
      cs = sprod(r1,r2);
      o1 = sprod(c1,r1); o2 = sprod(c2,r2);
      p1 = (o1-cs*o2)/(1-cs*cs); p2 = (o2-cs*o1)/(1-cs*cs);
      for(i=0; i<3; i++)
	(*s)->par[0][i] = p1*r1[i]+p2*r2[i];
      
      /* direction of line */
      vprod((*s)->par[1],r1,r2);

      /* local Minkowski functionals */
      (*s)->v[2] = .5*saferacos(cs) / (3*pi);
      (*s)->v[3] = 0;

      break;

    case nobody : break;
    }
    break;

  case nobody : break;
  }  

  /* set the parent pointers */
  (*s)->parent[0] = a;
  (*s)->parent[1] = b;
}
/* test for intersection */
boolean intersection(body*a,body*b)
{
  double d,ra,rb;

  if(a==b) return false;

  switch(a->type) {
  case ball : 
  case hole :
    switch(b->type) {
    case ball :		/* distance between sum and difference */ 
    case hole :		/* of radii */
      d = dist(a->par[0],b->par[0]); ra = a->par[1][0]; rb = b->par[1][0];
      return (ra+rb+epsilon>d && d>fabs(ra-rb)+epsilon);
    case wall :		/* radius greater than distance */
      return (a->par[1][0] > fabs(bodydist(b,a->par[0]))+epsilon);
    case nobody : return false; break;
    }
  case wall :
    switch(b->type) {
    case ball : 
    case hole :		/* same as above */
      return (b->par[1][0] > fabs(bodydist(a,b->par[0]))+epsilon);
    case wall :		/* normals not parallel */
      return (fabs(sprod(a->par[1],b->par[1]))+epsilon < 1);
    case nobody : return false; break;
    }
  case nobody :
    return false;
  }  

  return false;
}

/* ------------------------------------ */
/*                                      */
/* calculate scan points for object     */
/* surface element is taken as argument */
/* but may be adjusted                  */
/* ------------------------------------ */
double**scanpoints(body*b,int*nscan,double*ds)
{
  static double**u = (double**)NULL;
  static body*old = (body*)NULL;
  int ntheta,*nphi, i,j,k;
  double radius, theta,phi,dtheta,*dphi;
  int n[2];
  double x[2],dx[2],min[2],max[2], a[2][3], p[3],v[3], norm,d[3], lambda;

  /* the following intialization should correspond ds and nscan - */
  /* it does not see wall - check it */
  n[0]=20;  n[1]=20;
  dx[0]=1./(double)n[0];   dx[1]=1./(double)n[1];  

  if(!old) old = (body*)calloc(1,sizeof(body));

  switch(b->type) {
  case ball :
  case hole :
    /* try to reuse old values */
    if(u && old->type!=wall && b->type!=wall && old->par[1][0]==b->par[1][0] 
       && old->n<b->n) {
      for(i=0; i<*nscan; i++) 
	for(j=0; j<3; j++) u[i][j] += b->par[0][j]-old->par[0][j];
      old = b; return u;
    }
      
    /* theta is azimuth, phi is polar */
    /* points on equator */
    radius = b->par[1][0];
    /* ntheta = pi*radius/sqrt(*ds); dtheta = pi/ntheta; */
    ntheta = 100; dtheta = pi/ntheta;
    nphi = (int*)calloc(ntheta+1,sizeof(int));
    dphi = (double*)calloc(ntheta+1,sizeof(double));

    /* points at other latitudes */
    if(u) free_dmatrix(u,0,*nscan-1,0,2);
    for(theta=0, i=0, *nscan=0; i<=ntheta; theta+=dtheta, i++) {
      nphi[i] = (int)(2*ntheta*sin(theta))+1; dphi[i] = 2*pi/nphi[i];
      *nscan += nphi[i];
    }
    u = dmatrix(0,*nscan-1,0,2);

    /* calculate surface vectors */
    for(i=0, k=0, theta=0; i<=ntheta; theta+=dtheta, i++)
      for(j=0, phi=0; j<nphi[i]; j++, k++, phi+=dphi[i]) {
	u[k][0] = b->par[0][0] + radius*sin(theta)*cos(phi);
	u[k][1] = b->par[0][1] + radius*sin(theta)*sin(phi);
	u[k][2] = b->par[0][2] + radius*cos(theta);
      }
      
    /* finish */
    *ds = 4.*pi*radius*radius / (*nscan);
    free((void*)nphi); free((void*)dphi);
    old = b; return u;
    break;
    
  case wall : 
    /* without fully determining the geometry of our configuration, */
    /* we don't really know where to stop scanning a plane -- we'll */
    /* just stay inside the unit box and wait for a better idea ... */
    /* x and y are coordinates on plane */

    /* two vectors in plane */
    orthsystem(a[0],a[1],b->par[1]);
    for(i=0;i<2;i++) min[i]=infinity, max[i]=-infinity;
    
    /* calculate coordinates of each edge of cube intersecting plane */
    for(i=0;i<12;i++) {
      /* point on and direction vector of edge -- i/4 is index */
      p[i/4] = 0; p[(i/4+1)%3] = i%2; p[(i/4+2)%3] = i%4/2;
      v[i/4] = 1; v[(i/4+1)%3] = 0;   v[(i/4+2)%3] = 0;
      if(fabs(norm=sprod(v,b->par[1])) > epsilon) {
	/* coordinates on edge and in plane */
	dvec(d,p,b->par[0]); lambda = sprod(d,b->par[1])/norm;
	for(j = 0; j < 2; j++) {
	  x[j] = lambda*sprod(v,a[j])-sprod(d,a[j]);
	  if(x[j] < min[j])min[j] = x[j];
	  if(x[j] > max[j])max[j] = x[j];
	}
      }
    }

    /* calculate scan points */
    if(u) free_dmatrix(u,0,*nscan-1,0,2);
    *nscan = n[0]*n[1]; *ds = dx[0]*dx[1];
    u = dmatrix(0,*nscan-1,0,2);
    for(i=0, x[0]=min[0]; i<n[0]; i++, x[0]+=dx[0]) 
      for(j=0, x[1]=min[1]; j<n[1]; j++, x[1]+=dx[1])
	for(k = 0; k < 3; k++)
	  u[i*n[1]+j][k] = b->par[0][k] + x[0]*a[0][k] + x[1]*a[1][k];
    return u;
    break;

  case nobody : break;
  }

  return (double**)NULL;
}

/* --------------------------------------------- */
/*                                               */
/* a number of tests for points in a pattern     */
/* and related calculations                      */
/* --------------------------------------------- */
/* test if body b has any chance of reaching into window of pattern */
boolean irrelevant(body*b, pattern*pat)
{
  int i;

  switch(b->type) {
  case hole : return false; break;
  case wall : return false; break;
  case ball : 
    for(i=0; i<pat->nwin; i++)
      if(bodydist(&pat->b[i],b->par[0]) - b->par[1][0] > epsilon) return true;
    return false; break;
  case nobody : return false; break;
  }
  return false;
}
/* test if x lies in volume defined by pattern */
boolean involume(pattern*pat, double*x)
{
  int i;
  boolean flag;

  /* point must be inside all window parts ... */
  for(i=0, flag=true; i<pat->nwin && flag; i++)
    flag = inbody(&pat->b[i],x);

  /* ... and inside at least one ball */
  if(flag)
    for(i=pat->nwin, flag=false; i<pat->n && !flag; i++)
      flag = inbody(&pat->b[i],x);
  return flag;
}
/* test if triple point x of bodies b,c and d */
/* lies within volume defined by pattern */
/* assumes that a valid neighbour list has been set up for b */
/* in case of hole, all balls have to be checked anyway */
boolean ontriple(body*b,body*c,body*d, pattern*pat, double*x)
{
  int i;
  boolean flag;
  body*nb;

  /* point must be inside all window parts ... */
  for(i=0, flag=true; i<pat->nwin && flag; i++)
    if(i!=b->n && i!=c->n && i!=d->n) flag = inbody(&pat->b[i],x);

  /* ... and outside any other ball if on a ball */
  /* or inside at least one ball if on window parts only */
  if(flag) {
    if(b->type==ball || c->type==ball || d->type==ball) {
      for(nb=b->touch; nb && flag; nb=nb->touch)
	if(nb->type==ball && b!=nb && c!=nb && d!=nb) flag = outbody(nb,x);
    } else {
      switch(b->type) {
      case hole :
	for(i=pat->nwin, flag=false; i<pat->n && !flag; i++)
	  flag = inbody(&pat->b[i],x);
	break;
      case wall :
	for(nb=b->touch, flag=false; nb && !flag; nb=nb->touch)
	  if(nb->type==ball) flag = inbody(nb,x);
	break;
      case ball : break;
      case nobody : break;
      }
    }
  }
  return flag;
}
/* test if x on intersection curve of bodies b and c */
/* lies within volume defined by pattern */
/* assumes that a valid neighbour list has been set up for b */
boolean oncurve(body*b, body*c, pattern*pat, double*x)
{
  int i;
  boolean flag;
  body*nb;

  /* point must be inside all window parts ... */
  for(i=0, flag=true; i<pat->nwin && flag; i++)
    if(i!=b->n && i!=c->n) flag = inbody(&pat->b[i],x);

  /* ... and outside any other ball if on a ball */
  /* or inside at least one ball if on window parts only */
  if(flag) {
    if(b->type == ball || c->type == ball) {
      for(nb=b->touch; nb && flag; nb=nb->touch)
	if(nb->type==ball && b!=nb && c!=nb) flag = outbody(nb,x);
    }
    else
      switch(b->type) {
      case hole :
	for(i=pat->nwin, flag=false; i<pat->n && !flag; i++)
	  flag = inbody(&pat->b[i],x);
	break;
      case wall :
	for(nb=b->touch, flag=false; nb && !flag; nb=nb->touch)
	  if(nb->type==ball) flag = inbody(nb,x);
	break;
      case ball : break;
      case nobody : break;
      }
  }

  return flag;
}
/* test if x on surface of body b lies within volume defined by pattern */
/* assumes that a valid neighbour list has been set up for b */
boolean onsurface(body*b, pattern*pat, double*x)
{
  int i;
  boolean flag;
  body*nb;

  /* point must be inside all window parts ... */
  for(i=0, flag=true; i<pat->nwin && flag; i++)
    if(i != b->n) flag = inbody(&pat->b[i],x);

  /* ... and outside any other ball if on a ball */
  /* or inside at least one ball if on a window part */
  if(flag) {
    switch(b->type) {
    case ball :
      for(nb=b->touch; nb && flag; nb=nb->touch)
	if(nb->type==ball && b!=nb) flag = outbody(nb,x);
      break;
    case hole :
      for(i=pat->nwin, flag=false; i<pat->n && !flag; i++)
	flag = inbody(&pat->b[i],x);
      break;
    case wall :
      for(nb=b->touch, flag=false; nb && !flag; nb=nb->touch)
	if(nb->type==ball) flag = inbody(nb,x);
      break;
    case nobody : break;
    }
  }
  return flag;
}
/* test if x lies inside body b */
boolean inbody(body*b, double*x)
{
  if(bodydist(b,x) < epsilon) return true; else return false;
}
/* test if x lies outside body b */
boolean outbody(body*b, double*x)
{
  if(bodydist(b,x) > -epsilon) return true; else return false;
}
/* distance of x from body b; positive if outside */
double bodydist(body*b,double*x)
{
  double d[3];

  switch(b->type) {
  case ball :
  case hole : return dist(x,b->par[0])-b->par[1][0]; break;
  case wall : dvec(d,b->par[0],x); return sprod(d,b->par[1]); break;
  case nobody: return infinity; break;
  }
  return infinity;
}

/* --------------------------------------------- */
/*                                               */
/* some standard calculations involving vectors  */
/* --------------------------------------------- */
/* calculate two unit vectors orthogonal to u */
void orthsystem(double*n1,double*n2,double*u)
{
  double norm;

  if(fabs(u[0])<epsilon && fabs(u[1])<epsilon) {
    norm = safersqrt(u[0]*u[0]+u[2]*u[2]);
    n1[0] = -u[2]/norm; n1[1] = 0; n1[2] = u[0]/norm;
  }
  else {
    norm = safersqrt(u[0]*u[0]+u[1]*u[1]);
    n1[0] = -u[1]/norm; n1[1] = u[0]/norm; n1[2] = 0;
  }
  vprod(n2,n1,u);      
}
/* calculate vector product of two vectors u,v and normalize result */
void vprod(double*vp,double*u,double*v)
{
  int i;
  double d, max;

  vp[0] = u[1]*v[2]-u[2]*v[1];                     max=fabs(vp[0]);
  vp[1] = u[2]*v[0]-u[0]*v[2]; if(fabs(vp[1])>max) max=fabs(vp[1]);
  vp[2] = u[0]*v[1]-u[1]*v[0]; if(fabs(vp[2])>max) max=fabs(vp[2]);
  for(i=0; i<3; i++) vp[i] /= max;
  d = safersqrt(vp[0]*vp[0]+vp[1]*vp[1]+vp[2]*vp[2]);
  for(i=0; i<3; i++) vp[i] /= d;
}
/* determine distance of points u and v */
double dist(double*u, double*v)
{
  int i;
  double r, d = 0;
  
  for(i = 0; i < 3; i++) {
    r = v[i]-u[i];
    d += r*r;
  }
  return safersqrt(d);
}
/* return distance vector from u to v in r */
double dvec(double*r, double*u, double*v)
{
  int i;
  double d = 0;
  
  for(i = 0; i < 3; i++) {
    r[i] = v[i]-u[i];
    d += r[i]*r[i];
  }
  return safersqrt(d);
}

/* scalarproduct of two vectors */
double sprod(double*u, double*v)
{
  return u[0]*v[0]+u[1]*v[1]+u[2]*v[2];
}

/* ------------------------------------ */
/*                                      */
/* evaluation of functions              */
/* with test for sensible argument      */
/* ------------------------------------ */
double saferatan2(double x,double y)
{
  if(fabs(y)>epsilon) return atan2(x,y);
  if(x>0) return pi/2; else return -pi/2;
}
double saferacos(double x)
{
  if(x > 1)return 0; if(x < -1)return pi;
  return acos(x);
}
double saferasin(double x)
{
  if(x > 1)return pi/2; if(x < -1)return -pi/2;
  return asin(x);
}
double safersqrt(double x)
{
  if(x < epsilon)return 0;
  return sqrt(x);
}
 
void outputlist(triple*tr)
{
  triple*t;
  int j;

  if(tr) fprintf(stderr,"\nLIST");
  for(t=tr,j=0;t&&j<10;t=t->next,j++) {
    fprintf(stderr,"\ntriple %d %d %d at %10g",
	    t->parent[0]->n, t->parent[1]->n,
	    t->parent[2]->n, t->par);
    if(t->prev)
      fprintf(stderr," previous %d %d %d at %10g",
	      t->prev->parent[0]->n, t->prev->parent[1]->n, 
	      t->prev->parent[2]->n, t->prev->par);
    if(t->next)
      fprintf(stderr," next %d %d %d at %10g",
	      t->next->parent[0]->n, t->next->parent[1]->n, 
	      t->next->parent[2]->n, t->next->par);
  }
  if(tr) fprintf(stderr,"\nDONE");
}
