#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "types.h"
#include "rand.h"
#include "recipes.h"
#include "geometry.h"

#define pi 3.141592653589793238462643
#define infinity 6e66
#define epsilon 1e-10

/* Declarations */
/* calculation of Minkowski functionals */
double*minkowski(pattern*, parlist*, double);
void surfacecalc(pattern*, body*);
double chitheta(double,section*,body*,pattern*);
double montecarlo(pattern*,parlist*,double);

/* buffer zone and such */
void removeboundary(double*,double*);
double*buffer(pattern*,double,parlist*);

/* theoretical values */
double*theory(double, parlist*, double*);

/* ---------------------------------- */
/*                                    */
/*  main Minkowski measuring routine  */
/* ---------------------------------- */
double*minkowski(pattern*pat, parlist*par, double buf)
{
  double*v = (double*)calloc(4,sizeof(double));
  int i,j;
  body*b,*c,*d;
  static section*sect = (section*)NULL;
  triple*trip = (triple*)NULL;
  
  /* initialize bodies */
  for(i=0; i<pat->n; i++) {
    for(j=0; j<=3; j++) pat->b[i].v[j]=0;
    pat->b[i].wurscht = irrelevant(&pat->b[i], pat);
    /* if(i!=603 && i!=724 && i!=1398 && i>11) pat->b[i].wurscht = true; */
  }

  /* loop over all bodies */
  for(b=&pat->b[i=0]; i<pat->n; b=&pat->b[++i]) if(!b->wurscht) {
    /* set up the neighbour list */
    b->touch=(body*)NULL;
    for(c=&pat->b[j=0]; j<pat->n; c=&pat->b[++j])
      if(!c->wurscht && b!=c && intersection(b,c))
	c->touch=b->touch, b->touch=c;

    /* loop over all neighbours; measure each relevant intersection */
    for(c=b->touch; c; c=c->touch) if(b->n<c->n) {
      sectioncurve(&sect,b,c);
      for(d=b->touch;d;d=d->touch) if(d!=c) triplepoint(&trip,sect,d,pat);
      curvelength(sect,trip,pat);
      emptylist(&trip);
    }

    /* calculate surface */
    surfacecalc(pat,b);

  } /* end of loop over all bodies */

  /* determine global functionals */
  v[0] = montecarlo(pat,par,buf);
  for(i=1; i<=3; i++) for(j=0, v[i]=0; j<pat->n; j++) v[i] += pat->b[j].v[i];

  return v;
}

/* ------------------------------------------------------ */
/*                                                        */
/* calculate surface of a body sitting in a pattern       */
/* uses neighbour list of body                            */
/* ------------------------------------------------------ */
void surfacecalc(pattern*pat,body*b)
{
  double chi=0, x[3],y[3], d, u[3], rad=b->par[1][0];
  int i, sign;
  static section*s;
  
  switch(b->type) {
  case ball :
  case hole :
    if(b->touch) {		/* body has neighbours */
      if(b->touch->touch) {	/* two or more neighbours */
	chi = b->v[1]/(4*pi*rad*rad);
	while(chi<0) chi += 1; while(chi>1-epsilon) chi -= 1;
      }
      else {			/* a single neighbour */
	sectioncurve(&s,b,b->touch);
	d = dvec(u,s->par[0],b->par[0]); 
	if(d > epsilon) for(i=0; i<3; i++) u[i] /= d; 
	else vprod(u,s->par[1],s->par[2]);
	curvepoint(y,s,0); for(i=0; i<3; i++) y[i] -= b->par[0][i];
	for(sign=-1; sign<=+1; sign+=2) {
	  for(i=0; i<3; i++) x[i] = b->par[0][i] + sign * b->par[1][0]*u[i];
	  if(onsurface(b,pat,x)) chi = .5*(1-sign*sprod(u,y)/rad);
	}
      }
    }
    else {			/* no neighbours */
      for(i=0; i<3; i++) x[i]=b->par[0][i]; x[0] += rad;
      if(onsurface(b,pat,x)) chi=1; else chi=0;
    }

    /* evaluate contribution to functionals */
    if(chi) {
      b->v[1] = chi* 4*pi*rad*rad / 6;
      b->v[2] += chi* 4*pi*rad / (3*pi);
      b->v[3] += chi;
    }
    break;
  case wall : 
    /* surface area of walls has already been calculated by triangulation! */
    break;
  case nobody : 
    break;
  }
}
/* actually, we're integrating this function */
/* in the range theta from 0 to pi -- the section* is a dummy */
double chitheta(double theta, section*dummy,body*b,pattern*pat)
{
  int i;
  double rad, chi;
  static section*s=(section*)NULL;
  body*c;
  triple*t=(triple*)NULL;

  /* allocate memory for test circle */
  if(!s) {
    s = (section*)calloc(1,sizeof(section));
    s->type = circ;
    s->par = (double**)calloc(4,sizeof(double*));
    for(i=0; i<3; i++) s->par[i] = (double*)calloc(3,sizeof(double));
    s->parent = (body**)calloc(2,sizeof(body*));
  }

  /* initialize test circle */
  rad = b->par[1][0]*sin(theta);
  for(i=0; i<3; i++) {
    s->par[0][i] = b->par[0][i] + (i==2)*b->par[1][0]*cos(theta);
    s->par[1][i] = (i==0)*rad;
    s->par[2][i] = (i==1)*rad;
  }
  s->parent[0] = b; s->parent[1] = b;
  
  /* evaluate contribution to chi */
  for(c=b->touch; c; c=c->touch) triplepoint(&t, s,c,pat);
  chi = sin(theta)*curvelength(s,t,pat)/(4*pi);

  /* sod off */
  emptylist(&t);
  return chi;
}
/* ---------------------------------------------- */
/*                                                */
/* Monte Carlo integration of volume              */
/* enclosed by some objects                       */
/*                                                */
/* points are thrown into the (buffered) cube     */
/* ---------------------------------------------- */
double montecarlo(pattern*pat, parlist*par, double buf)
{
  int imin=10000, imax = 100000, i,j, volume=0;
  double err=1., errmax = 1.e-5, x[3], lo[3],hi[3],box;
  double used_buf;

  /* do we use a buffer? */
  if(par->rmbound) 
    used_buf = buf;
  else
    used_buf = 0.;

  for(i=0; i<3; i++) lo[i]=used_buf, hi[i]=1-used_buf;
  /* speedup for clusters */
  switch(par->hemisphere) {
  case 1 : lo[2] = .5+used_buf; break;
  case 2 : hi[2] = .5-used_buf; break;
  }
  /* volume of (buffered) cube */
  for(i=0, box=1; i<3; i++) {
    if(hi[i]-lo[i]<epsilon) return 0;
    box *= hi[i]-lo[i];
  }

  /* throw Monte Carlo points up to maximum number or until error is small */
  for(i=0; i<imax && err>errmax; i++) {
    for(j=0; j<3; j++) x[j] = (hi[j]-lo[j])*rand1()+lo[j];
    if(involume(pat,x)) volume++;

    /* the (running) error estimate, */
    /* this can be dangerous if imin is to small */
    if(i<imin || volume==0) err = 1; 
    else err = 1./volume - 1./i;
  }

  return box*(double)volume/i;
}

/* ------------------------------------ */
/*                                      */
/* remove the boundary contribution vd  */
/* from functionals v                   */
/* ------------------------------------ */
/* binomial coefficient */
int binko(int top, int bottom)
{
  if(top == 1 || bottom == top || bottom == 0) return 1;
  else return binko(top-1,bottom)+binko(top-1,bottom-1);
}
void removeboundary(double*v,double*vd)
{
  int i,j;
  double m[4],md[4];

  m[0] =          v[0]; md[0] =          vd[0];
  m[1] = .75    * v[1]; md[1] = .75    * vd[1];
  m[2] = 1.5/pi * v[2]; md[2] = 1.5/pi * vd[2];
  m[3] = .75/pi * v[3]; md[3] = .75/pi * vd[3];

  for(i = 0; i <= 3; i++)
    {
      for(j = 0; j < i; j++)
	m[i] -= binko(i,j)*m[j]*md[i-j];
      m[i] /= md[0];
    }

  /* change normalization from m to v */
  v[0] = m[0];
  v[1] = m[1] / .75;
  v[2] = m[2] * pi/1.5;
  v[3] = m[3] * pi/.75;
}

/* ------------------------------------------------------ */
/*                                                        */
/* create the window part of a pattern by pushing the     */
/* original window boundaries inward                      */
/* if this leaves anything, return Minkowski functionals  */
/* otherwise null pointer                                 */
/* ------------------------------------------------------ */
double*buffer(pattern*pat,double buf,parlist*par)
{
  int i,j;
  boolean left=true;
  double*v;

  /* copy window to pattern and shrink according to buffer size */
  pat->b = (body*)realloc(pat->b,(pat->n = pat->nwin+1)*sizeof(body));
  for(i=0; i<pat->nwin && left; i++) {
    pat->b[i].n = i;
    switch(pat->win[i].type) {
    case hole : 
      addhole(&pat->b[i], pat->win[i].par[0], pat->win[i].par[1][0]);
      if((pat->b[i].par[1][0] -= buf) < epsilon) left=false;
      break;
    case wall : 
      addwall(&pat->b[i], pat->win[i].par[0], pat->win[i].par[1]);
      for(j=0; j<3; j++) pat->b[i].par[0][j] -= buf*pat->b[i].par[1][j];
      break;
    case ball : 
      addball(&pat->b[i], pat->win[i].par[0], pat->win[i].par[1]);
      if((pat->b[i].par[1][0] -= buf) < epsilon) pat->b[i].par[1][0] = 0;
      break;
    case nobody : 
      break;
    }
  }

  /* add one giant ball to calculate Minkowski functionals */
  makeball(&pat->b[pat->nwin], .5,.5,.5, 1.); pat->b[pat->nwin].n = pat->nwin;
  v = minkowski(pat,par,buf);
  pat->b = (body*)realloc(pat->b,--pat->n * sizeof(body));

  /* if window is left, return functionals */
  if(left && v[0] > epsilon) return v; else return (double*)NULL;
}

/* --------------------------------------------------- */
/*                                                     */
/* calculate Minkowski functionals for Poisson process */
/* --------------------------------------------------- */
double*theory(double radius, parlist*par, double*vd)
{
  int d=3, i,j, n = par->n;
  static double*v = (double*)NULL;
  double ms[4], mn[4];
  double unit[4], uncvr, omega, rho, sqr, cbc;
  
  /* volume of unit balls */
  unit[0] = 1.;
  unit[1] = 2.;
  unit[2] = pi;
  unit[3] = 4.*pi/3.;

  /* volume of whole sample */
  omega = vd[0];
  
  /* Minkowski functionals for single ball */
  ms[0] = unit[d-0]/unit[0]* radius*radius*radius;
  ms[1] = unit[d-1]/unit[1]* radius*radius;
  ms[2] = unit[d-2]/unit[2]* radius;
  ms[3] = unit[d-3]/unit[3];
  
  /* some useful values */
  /* uncovered volume */
  uncvr = 1; for(i = 1; i <= n; i++)uncvr *= 1-ms[0]/omega;
  /* Density and its powers for finite volume and number */
  rho = n/(omega-ms[0]);
  sqr = rho*(n-1)/(omega-ms[0]);
  cbc = sqr*(n-2)/(omega-ms[0]);
  
  /* Minkowski functionals for n balls */
  mn[0] = 1-uncvr;
  mn[1] = uncvr* rho*ms[1];
  mn[2] = uncvr*(rho*ms[2] -   sqr*ms[1]*ms[1]);
  mn[3] = uncvr*(rho*ms[3] - 3*sqr*ms[1]*ms[2] + cbc*ms[1]*ms[1]*ms[1]);
  
  /* normalization */
  if(!v) v=(double*)calloc(4,sizeof(double));
  for(j=0; j<=3; j++) v[j]=unit[d]/unit[d-j]*mn[j];
  return v;
}
